"""
Module provinding a function that estimates a model's input/output tensor names.
"""

from tfserve.loader import load_model


out_hints = ["softmax", "sigmoid", "out", "output", "prediction",
             "probability", "prob", "inference"]


def estimate_io_tensors(model_path):
    """
    Prints estimates input/output tensor names that could be used later
    with the TFServeApp class to serve the model.

    :param str model_path: can be a '.pb' model file or a checkpoint directory.
    """
    sess = load_model(model_path)
    graph = sess.graph

    print("Possible INPUT tensor names:")
    ph = [op.outputs[0] for op in graph.get_operations() if op.type == "Placeholder"]
    for p in ph:
        print("\t{}".format(p.name))

    print()
    print("Possible OUTPUT tensor names:")
    ops = [n for n in graph.get_operations()]

    for op in ops:
        if any(x in op.name.lower() for x in out_hints):
            for o in op.outputs:
                print("\t{}".format(o.name))

    sess.close()



import os
import sys

def _check_tensorflow():
    """Check tensorflow status.

    Premptively import tensorflow and exit with an error message if
    it's not installed.

    """
    try:
        import tensorflow
    except ImportError:
        sys.stderr.write(
            "tfserve: tensorflow is not installed\n"
            "Try installing it by running 'pip install tensorflow' or\n"
            "'pip install tensorflow-gpu' if your system supports GPUs.\n"
            "For more information see https://www.tensorflow.org/install/\n")
        sys.exit(1)

_check_tensorflow()

from tfserve.tfserve import TFServeApp
from tfserve.tfserve import BadInput



"""
Auxiliary methods that operate on tensorflow graphs.
This are private and should not be used by users.
"""

import re


def check_placeholders(graph, tensors):
    """
    Check that tensors are valid graph placeholders.
    """

    if graph is None or tensors is None:
        raise ValueError("None of the parameters can be None")

    ph = [op for op in graph.get_operations() if op.type == "Placeholder"]
    ph_names = [p.outputs[0].name for p in ph]

    for t in tensors:
        t = smart_tensor_name(t)
        if t not in ph_names:
            raise ValueError("Input node must be placeholder: {}".format(t))


def check_tensors(graph, tensors):
    """
    Check that tensors are valid graph tensors.
    """

    if graph is None or tensors is None:
        raise ValueError("None of the parameters can be None")

    for t in tensors:
        try:
            graph.get_tensor_by_name(smart_tensor_name(t))
        except (ValueError, KeyError):
            raise ValueError("Non existent tensor in graph: {}".format(t))


def check_input(provided, required, msg):
    """
    Check that provided and required set of tensors are equal.
    If not, raises ValueError with msg
    """
    if provided is None or required is None:
        raise ValueError("None of the parameters can be None")

    for x in provided:
        if x not in required:
            raise ValueError(msg)

    for x in required:
        if x not in provided:
            raise ValueError(msg)


def smart_tensor_name(t):
    """
    For check_placeholders and check_tensors functions, the tensors
    can either be the correct tensor name (finishing with :digit) or
    operation names. If user provided operations (common error in tensorflow),
    the method will assing :0 to each of them to get the tensor name.
    """
    m = re.search(r':\d+$', t)
    if m is None:
        return "{}:0".format(t)
    return t



"""
Module that handles loading a tensorflow model in several different forms.
"""
import os
import tensorflow as tf


def load_model(model_path):
    """
    Loads a tensorflow model return a tf.Session running on the loaded model (the graph).

    :param str model_path: It can be a `.pb` file or directory containing checkpoint files.

    :return: tf.Session running the model graph.
    """
    if model_path is None:
        raise ValueError("model_path must not be None")

    if not os.path.exists(model_path):
        raise ValueError("model_path must exist")

    if os.path.isfile(model_path) and model_path.endswith(".pb"):
        return _load_pb(model_path)

    if os.path.isdir(model_path):
        for f in os.listdir(model_path):
            if f.endswith(".pb"):
                return _load_pb(os.path.join(model_path, f))

        return _load_ckpt(model_path)


def _load_pb(model_path):
    """
    Loads from a '.pb' model file.
    """
    graph = tf.Graph()
    sess = tf.Session(graph=graph)
    with tf.gfile.FastGFile(model_path, 'rb') as f:
        graph_def = tf.GraphDef()
        graph_def.ParseFromString(f.read())
        with sess.graph.as_default():
            tf.import_graph_def(graph_def)

    return sess


def _load_ckpt(model_dir):
    """
    Loads from a checkpoint directory.
    """
    graph = tf.Graph()
    sess = tf.Session(graph=graph)
    with graph.as_default():
        ckpt_path = tf.train.latest_checkpoint(model_dir)
        saver = tf.train.import_meta_graph('{}.meta'.format(ckpt_path))
        saver.restore(sess, ckpt_path)
        return sess



import json

import numpy as np
from apistar import http, App, Route

from werkzeug import routing
from werkzeug import serving
from werkzeug.exceptions import BadRequest
from werkzeug.exceptions import HTTPException
from werkzeug.exceptions import MethodNotAllowed
from werkzeug.wrappers import Request, Response

from tfserve.loader import load_model
import tfserve.graph_utils as graph_utils


class BadInput(Exception):
    """
    Raised by decode when input is invalid.
    """

    def __init__(self, description):
        super(BadInput, self).__init__(description)
        self.description = description


class TFServeApp():
    """
    This class will handle all server functionality for one-to-one models.
    A TFServeApp object is able to run a web server to serve the model and
    will also be responsible for encoding and decoding it's input/output.

    Each TFServeApp will be running it's own tensorflow session and graph.

    The user only needs to create the objects giving the appropiate parameters
    and execute the `run` method that loads everything up. From that moment,
    the server will be up and listening for input data.


    Examples of user use:

    app = TFServeApp("logistic_regression_model.pb", ["input:0"], ["sigmoid:0"]
                        lambda x: {"input:0": np.array(x)},
                        lambda y: {"prob": double(y["sigmoid:0"])})
    app.run("127.0.0.1", 5000)

    A web server should be up and running, being POST to / it's only supported method.
    In this request, the model input should be provided as the request body data. This data
    will be received by the encode function that will process it to feed the model. The server will
    run the model, giving it's output to the decode function that will prepare the reponse data.
    """

    def __init__(self, model_path, in_t, out_t, encode, decode, batch=False):
        """
        When constructing, the method checks that all in_t tensors are valid
        placeholders and all out_t tensors are valid tensors that exist in
        the graph.

        The objects is constructed by providing:
            * A model path
            * List of input placeholder that will be used to feed the model.
            * List of expected outputs tensors (the desired predictions).
            * encode function mapping request data to input numpy values.
            * decode function mapping output numpy values to request response data.

        :param str model_path: It can be a `.pb` file or directory containing checkpoint files.
        :param list[str] in_t: List of placeholder tensor names. Something like: ["input/image:0"]
        :param list[str] out_t: List of output tensor names. Something like: ["output/Softmax:0"]
        :param encode: python function that receives the request body data and returns a `dict` mapping
                        in_t to numpy values.
        :param decode: python function that receives a `dict` mapping out_t to numpy values and returns
                        the response data (for example, a `dict` object that will be transformed to JSON).
                        The return object of this method will be the response to the request.
                        Read it's docs for more information on how to return certain objects (for example, images).
        :param list[str] out_t: List of output tensor names. Something like: ["output/Softmax:0"]
        :param boolean batch: If False, batch dimension (required by tensorflow) will be automatically
                               handled (that is, you don't need to handle it yourself in the encode/decode functions).
                               This option is ideal when dealing with single inferences.
                               If True, you can run multiple inferences at the same time by dealing
                               with the batch dimension yourself in the encode/decode functions.

        :raises ValueError: if in_t are not all placeholder or out_t contains non-existent graph tensors
        """
        self.sess = load_model(model_path)
        self.graph = self.sess.graph

        graph_utils.check_tensors(self.graph, in_t)
        graph_utils.check_tensors(self.graph, out_t)

        graph_utils.check_placeholders(self.graph, in_t)

        self.in_t = in_t
        self.out_t = out_t

        self.encode = encode
        self.decode = decode

        self.batch = batch

    def _make_inference(self, request: http.Request):
        """
        This method is the request handler. It deals with the logic of encoding the input, running the model
        and decoding it's output for final response.

        When encoding the input, it also makes sure that all the encode function creates numpy values for all
        in_t placeholders.

        :raises ValueError: if the encoded data provided by the request does not include all
                            the in_t placeholders.
        """
        return self._make_inference_impl(request.body)

    def _make_inference_impl(self, req_bytes):
        """
        Implementation of _make_inference to decouple data interface from HTTP server interface.

        req_bytes must be the unencoded HTTP request body as bytes.

        Returns a Python dict that can be encoded as a JSON HTTP response.
        """
        feed_dict = self.encode(req_bytes)
        if not self.batch:
            feed_dict = {k: np.expand_dims(v, axis=0) for k, v in feed_dict.items()}

        feed_dict = {graph_utils.smart_tensor_name(k): v for k, v in feed_dict.items()}


        graph_utils.check_input(feed_dict.keys(), self.in_t, "Encode function must generate all and only input tensors")

        out_map = {}
        ret = self.sess.run(self.out_t, feed_dict=feed_dict)
        for i, e in enumerate(self.out_t):
            out_map[e] = ret[i] if self.batch else np.squeeze(ret[i])

        return self.decode(out_map)

    def run(self, host, port, middleware=None):
        """Werkzeug run implementation.

        `middleware` may be provided as a function to handle
        requests. It must accept the arguments `(handler, req)` where
        `handler` is the TFServeApp request handler and `req` is the
        request.

        """
        app = self._init_app(middleware)
        server = serving.make_server(
            host, port, app, threaded=True,
            request_handler=serving.WSGIRequestHandler)
        server.serve_forever()

    def _init_app(self, middleware=None):
        """Initialize a WSGI application for handling POST to '/'.

        `middleware` may be provided as WSGI middleware.

        """
        routes = routing.Map([
            routing.Rule('/', endpoint=self._handle_inference),
            routing.Rule('/ping', endpoint=self._handle_ping),
            routing.Rule('/shutdown', endpoint=self._handle_shutdown),
        ])
        def app(env, start_resp):
            """WSGI application to handle server requests.

            """
            urls = routes.bind_to_environ(env)
            try:
                handler, _kw = urls.match()
                req = Request(env)
                if middleware:
                    return middleware(handler, req)(env, start_resp)
                return handler(req)(env, start_resp)
            except HTTPException as e:
                return e(env, start_resp)
        return app

    def _handle_inference(self, req):
        """Handle inference request.

        """
        if req.method != 'POST':
            raise MethodNotAllowed(valid_methods=['POST'])
        req_bytes = req.stream.read()
        try:
            resp_val = self._make_inference_impl(req_bytes)
        except BadInput as e:
            raise BadRequest(e.description)
        return Response(json.dumps(resp_val), content_type='application/json')

    @staticmethod
    def _handle_ping(_req):
        """Handles ping request.

        """
        return Response()

    @staticmethod
    def _handle_shutdown(req):
        """Handles shutown request.

        """
        if req.method != 'POST':
            raise MethodNotAllowed(valid_methods=['POST'])
        shutdown = req.environ.get('werkzeug.server.shutdown')
        if not shutdown:
            raise BadRequest("server does not support shutdown")
        shutdown()
        return Response()

    def run_apistar(self, *args, **kwargs):
        """
        LEGACY CODE THAT RUNS UNDER APISTAR

        After building the object, this method needs to be called to run the server. This method
        will build the apistar `app` with the appropiate handlers and will serve it.

        It receives the same parameters as apistar App.serve method. More information here:
        https://github.com/encode/apistar/blob/master/apistar/server/app.py#L161

        :raises ValueError: if the encoded data provided by the request does not include all
                            the in_t placeholders.
        """
        routes = [Route('/', method='POST', handler=self._make_inference)]

        app = App(routes=routes)
        app.serve(*args, **kwargs)




#!/usr/bin/env python

import sys
from setuptools import setup

if sys.version_info < (3, 5):
    raise NotImplementedError("Sorry, you need at least Python 3.5 to use tfserve.")

def readme():
    with open('README.md') as f:
        return f.read()

with open('requirements.txt') as f:
    required = f.read().splitlines()

setup(name='tfserve',
      version="0.2.1",
      description='Serve TF models simple and easy as an HTTP API server.',
      long_description=readme(),
      long_description_content_type="text/markdown",
      author="Ivan Itzcovich",
      author_email='i.itzcovich@gmail.com',
      url='http://github.com/iitzco/tfserve',
      keywords='tensorflow deep-learning serving',
      # scripts=["bin/tfserve"],
      packages=['tfserve'],
      license='MIT',
      platforms='any',
      install_requires=required,
      python_requires='>3.5',
      classifiers=['License :: OSI Approved :: MIT License',
                   'Topic :: Internet :: WWW/HTTP :: HTTP Servers',
                   'Topic :: Internet :: WWW/HTTP :: WSGI',
                   'Topic :: Internet :: WWW/HTTP :: WSGI :: Application',
                   'Topic :: Internet :: WWW/HTTP :: WSGI :: Middleware',
                   'Topic :: Internet :: WWW/HTTP :: WSGI :: Server',
                   'Topic :: Software Development :: Libraries :: Application Frameworks',
                   'Programming Language :: Python :: 3.5',
                   'Programming Language :: Python :: 3.6',
                   'Programming Language :: Python :: 3.7',
                   ],
      test_suite='nose.collector',
      tests_require=['nose'],
      include_package_data=True,
)
